package provide alicq 0.8.9

package require helper

# Modules handling code '''1
namespace eval modules {}

namespace eval module {
	set path  [list]
	array set cache [list]
}

# Module search path
proc module::path {args} {
	variable path
	variable cache
	if {[llength $args]} {
		foreach x $args { set path [concat $path $x] }
		array set cache [list]
		foreach item $path {
			foreach x [glob -nocomplain $item/*.tcl] {
				set name [file tail [file root $x]]
				if {[info exists cache($name)]} continue
				set cache($name) $x
			}
		}
	}
	set path
}

proc module::name {} {
	set ns [uplevel 1 namespace current]
	while {[namespace parent $ns]!="::modules"} {
		set ns [namespace parent $ns]
	}
	namespace tail $ns
}

# Get names of all available modules
proc module::names {{mask *}} {
	variable cache
	array names cache $mask
}

# Get names of all currently loaded modules
proc module::loaded {} {
	set res [list]
	foreach x [namespace children ::modules] {
		lappend res [namespace tail $x]
	}
	set res
}

# Get namespace of module name
proc module::space {name} { return "::modules::$name" }

proc meta2var {x} {
	return [namespace qualifiers [namespace qualifiers $x]]::[namespace tail $x]
}
proc module::init {ns} {
	set name [namespace tail $ns]
	# Initialize metavariables
	foreach x [_meta $ns] {
		upvar #0 $x meta
		if {![array exists $x]} continue
		set var [meta2var $x]
		foreach {key val} [array get $x] {
			set cmd [namespace current]::initmeta::$key
			if {[info commands $cmd]==$cmd} { $cmd $var $val }
		}
		after idle [list trace variable $var w module::__saver]
	}	
	variable cache
	if {[origin $name]==""} { set cache($name) "" }
}

proc module::belongto {var} {
	lindex [split [string map {:: "\0" } $var] "\0"] 2
}

proc module::__saver {var field op} {
	variable forsave
	set name [belongto $var]
	if {[info exists [space $name]::meta::_auto]} {
		set skip [info exists forsave]
		if {!$skip || [lsearch -exact $forsave $name] == -1} {
			lappend forsave $name
		}
		if {!$skip} { after idle module::__save }
	}
}

proc module::__save {} {
	variable forsave
	set mod [lindex $forsave 0]
	set forsave [lrange $forsave 1 end]
	# Saving itself
	set res "# This file was generated automatilcally, do not edit\n\n"
	foreach x [meta $mod] {
		if {![array exists $x]} continue
		set var [meta2var $x]
		if {[info exists $var] &&
		    (![info exists ${x}(default)] ||
		    [set ${x}(default)]!=[set $var])} {
			set name [join [lrange [split [string map {:: "\0"} $var] "\0"] 3 end] ::]
			append res [list set $name [set $var]] "\n"
		}
	}
	if {[catch {	set f [open [set [space $mod]::meta::_auto] w]
			puts -nonewline $f $res
			close $f	} reason]} { 
		puts "Error: $reason"
	}
	# Shedule saving of other files
	if {![llength $forsave]} { unset forsave} else {
		after idle module::__save
	}	
}

namespace eval module::initmeta {
	proc default {name val} {
		trace variable $name r [list\
			[namespace current]::do::default $val $name]
	}
	namespace eval do {
		proc default {default var name1 name2 op} {
			upvar #0 $var x
			if {![info exists x]} { set x $default }
			trace vdelete $var r [list\
				[namespace current]::default $default $var]
		}
	}
}

# Load module name unless it is already loaded
proc module::load {name} {
	if {[lsearch -exact [loaded] $name]!=-1} { return 0 }
	set space [space $name]
	set file [origin $name]
	if {$file==""} { return -code error "module $name not found"}
	if {[catch { namespace eval $space\
		[list source [origin $name]]} reason]} {
			catch { namespace delete $space }
			return -code error \
				"Error loading module $name: $reason"
	}
	variable autobase
	if {[info exists autobase]} {
		set path [file join $autobase $name]
		if {[file exists $path] &&
		    ![catch { namespace eval $space [list source $path] }]} {
		    	namespace eval ${space}::meta [list set _auto $path]
		}
	}
	init $space
	Event ModuleLoaded $name 
	return 1
}

# Unload module if it is possible
proc module::unload {name} {
	set ns [space $name]
	set cmd ${ns}::meta::unloadable
	if {[info commands $cmd]==$cmd && [string is false [$cmd]]}  {
		return -code error "Module $name can not be unloaded"
	}
	# Invoke module removal code if needed
	set cmd ${ns}::meta::unload
	if {[info commands $cmd]==$cmd} $cmd
	# Clean any Alicq event handlers pointing to module namespace
	set ns [space $name]
	# Delete traces from objects
	foreach obj [select] {
		set ref [ref $obj]
		foreach item [trace vinfo $ref] {
			foreach {type script} $item break
			if {[string match "${ns}::*" $script] ||
			    [string match "::namespace inscope *" $script] &&
			    [lindex $script 2]==$ns} { 
			    	trace vdelete $ref $type $script
			}
		}
	}
	# Destroy possible traces
	__untrace $ns
	# Destroy module namespace
	namespace delete ${ns}
	# Notify modules that module was unloaded
	catch { disable $name }
	Event ModuleUnloaded $name
}

proc module::__untrace {ns} {
	foreach x [info vars ${ns}::*] {
		foreach x [trace vinfo $x] {
			foreach {ops cmd} $x break
			trace vdelete $x $ops $cmd
		}
	}
	foreach x [namespace children $ns] { __untrace $x }
}

# Get name of the file the module is located in
proc module::origin {name} {
	variable cache
	if {![info exists cache($name)]} {
		return ""
	} 
	set cache($name)
}

# Get metainformation about all modules or given module's namespace
# configurable parameters
proc module::meta {name} { _meta [space $name] }
proc module::_meta {root} {
	set list [list]
	foreach x [info vars ${root}::meta::*] { lappend list $x }
	foreach x [namespace children $root] { 
		switch -- [namespace tail $x] meta continue
		set list [concat $list [_meta $x]]
	}
	set list
}

proc module::enable {name} {
	variable autobase
	if {![info exists autobase]} return
	if {![file isdirectory $autobase]} { file mkdir $autobase }
	
	set path [file join $autobase disabled $name]
	set config [file join $autobase $name]
	if {[file exists $path]} {
		file rename $path $config
		catch { namespace eval [space $name] [list source $config]}
	} else {
		catch { set f [open $config w]; close $f }
	}
	namespace eval [space $name]::meta [list set _auto $config]
}

proc module::disable {name} {
	variable autobase
	if {![info exists autobase] || ![file isdirectory $autobase]} return
	
	set path [file join $autobase disabled]
	if {![file isdirectory $path]} { file mkdir $path }
	if {[file exists [file join $autobase $name]]} {
		set new [file join $path $name]
		if {[file exists $new]} { file delete $new }
		file rename [file join $autobase $name] $new
	}
}

proc module::autoload {base} {
	variable autobase
	set autobase $base
	foreach path [glob -nocomplain -type f [file join $base *]] {
		set name [file tail $path]
		if {[info exists [space $name]::meta::_auto]} continue
		puts "Auto load $name, invoke $path"
		if {![catch { load $name } reason]} {
			set space [space $name]
			if {[catch {
				namespace eval $space [list source $path]
				} reason]} { Event Log error\
					"Module init: $reason" 
			} else { namespace eval ${space}::meta\
					[list set _auto $path] }
		} else { Event Log error $reason }
	}
}

# Return name of variable corresponding given module property name
proc module::var {prop} {
	return [space [lindex $prop 0]]::[join [lrange $prop 1 end] ::]
}
#```1

# Hook management commands '''1
proc hook {events script {priority 0.5} {filter 0}} {
	if {$priority<0 || $priority>1} {
		return -code error "Priority should be in the range 0-1"
	} else {set priority [expr $priority]}
	if {![string is boolean $filter]} {
		return -code error "Filter should be boolean"
	} else { set filter [string is true $filter]}
	set ns [uplevel 1 {namespace current}]
	set id [interp alias {} ${ns}::[_name $ns\
		[clock seconds]-[clock clicks] $priority $filter] {} $script]
	_hook $events $ns $id
}

proc _name {ns id priority filter} {
	list _hook $priority [namespace tail $ns] $id $filter
}

proc _hook {events ns id} {
	if {![info exists ${ns}::__hooked]} {
		namespace eval ${ns} { namespace export _hook* }
		set ${ns}::__hooked 1
	}
	foreach x $events {
		namespace eval ::hooks::${x} [list namespace import $id]
	}
	set id
}

proc hooklist {event} {
	set res [info commands ::hooks::${event}::*]
	foreach x [namespace children ::hooks {*\**}] {
		if {[string match [namespace tail $x] $event]} {
			set res [concat $res [info commands ${x}::*]]
		}
	}
	set res
}

proc unhook {id} {
	foreach x [info commands $id] { rename $x {} } 
}

proc Event {name args} {
	set res ""
	
	foreach x [lsort -index 1 -real [hooklist $name]] {
		set r [catch { set res [eval [list $x] $args] } v]
		if {[lsearch {0 2 3 4} $r]==-1} {
			if {$name!="Error:Module"} {
				Event Error:Module $x $v
			}
			if {$name!="Log"} {
				Event Log 0 "Error in $x: $v\n$::errorInfo" 
			}	
		} elseif {[lindex $x end]} {
			if {[llength $res]==[llength $args]} {
				set args $res
			} else {
				Event Error:Module $x\
				   "Filter returned wrong number of arguments"
			}
		}
		switch -exact -- $r {
			0 -
			4 continue 
			100 { rename $x {} }
			default break
		}
	}

	set res
}

proc _handler {filter events name arglist body {priority 0.5}} {
	set ns [uplevel 1 namespace current]
	set id ${ns}::[_name $ns $name $priority $filter]
	proc  $id $arglist $body
	_hook $events $ns $id
}
interp alias {} handler {} _handler 0
interp alias {} filter {} _handler 1

set flashcnt 0
proc Flash {type} {
	switch -- $type {
		more {if {[incr ::flashcnt]==1} {Flasher}}
		less {if {![incr ::flashcnt -1]} {
				after cancel $::flasher
				Event Flashing 1
			}	
		}
		* {return -code error "Unknown parameter $type for Flash"}
	}
}

proc Flasher {{stage 0}} {
	Event Flashing $stage
	set ::flasher [after 300 "Flasher [expr $stage^1]"]
}
#```1

set default_modules {icq history password notify tree dialogs config}
# Core modules, provided by alicq itself.
namespace eval modules {
	# Global actions
	namespace eval my::meta::menu {
		interp alias {} [namespace current]::[list Add Group] {}\
			Event AddItem Group:common
		interp alias {} [namespace current]::[list 0.99 Exit] {}\
			Event exit
	}
}
# Proxy servers management
namespace eval modules::Proxy {
	namespace eval meta {
		set name Proxies
		set description "Alicq core - proxies"
		set author "Ihar Viarheichyk <iverg@mail.ru>"
		set configuration-objects Proxy
	}
	package require proxy 0.3
	handler New:Proxy:* ProxyAdded {uid} {
		set ref [ref $uid]
		set name [lindex [split $uid :] end]
		foreach {key val} [array get $ref] { 
			proxy::proxy $name -$key $val
		}
		trace variable $ref w [nc ProxyChanged $name]
		trace variable $ref u [nc ProxyDeleted $name]
	}
	proc ProxyChanged {name ref field args} {
		upvar 1 ${ref}($field) value
		proxy::proxy $name -$field $value
	}
	proc ProxyDeleted {name ref field args} {
		if {$field!=""} return
		proxy::delete $name
	}
	module::init [namespace current]
}

namespace eval modules::log {
	array set mapping {error 0 warning 1 notice 3 info 5 debug 10}
	namespace eval meta {
		set name Logging
		set description "Alicq core - event logging"
		set author "Ihar Viarheichyk <iverg@mail.ru>"
		array set level { type variant save change }
		set level(values) [array names [namespace parent]::mapping]
	}
	set level error
	variable fd stderr

	proc loglevel {str} {
		variable mapping
		if {![string is integer $str]} {
			if {[info exists mapping($str)]} {
				set str $mapping($str)
			} else { set str 5 }
		}
		set str
	}

	proc init {name} {
		variable fd
		if {[catch {
			if {[file exists $name]} {file delete $name}
			set fd [open $name a+]
			fconfigure $fd -buffering none
		} reason]} {
			set fd stderr
			Event Log error "Can not open $name, using stderr: $reason"
		}
		Event Log 0 "Alicq started"
	}

	handler Log Log {tags str} {
		variable level
		variable fd
		set l [lindex $tags 0]
		if {[loglevel $l] > [loglevel $level]} return
		if {[llength $tags]} { set tags "\[$tags\]" }
		puts $fd [format "%s: %s %s" [clock format [clock seconds]\
				-format {%x %X}] $tags $str]
	}
	module::init [namespace current]
}
#interp alias {} log {} moduleconfig log

# Namespace 'objects' contains all objects which should be managed with 
# alicq helper procedures.
namespace eval objects {

	# Contacts of any type
	namespace eval Contact::meta {
		array set Alias {default "" save change}
		array set Groups {type groups default {} save change}
	}
	# Groups of any type
	namespace eval Group {
		# Common groups - permanent non-conditional membership
		namespace eval common::meta {
			array set Send:text {type action menu {Send Multicast} weight .10}
			array set Alias {default "" save change}
			array set Groups {type groups default {} save change}
		}
	}
	# Proxy servers
	namespace eval Proxy::meta {
		array set type {type variant weight .1 default http
			valuescript {proxy::types} save change}
		array set server {type fqdn save change weight .2}
		array set port {type {port digit} save change weight .3}
		array set user {save change weight .4}
		array set password {type password save change weight .5}
	}

	# Pseudo-object 'Me' to keep properties of current users similar to 
	# properties of ordinary users. These properties does not require 
	# metainformation, as user have no access to them.
	array set Me { Alias Me Status offline }
}

# Validation commands
namespace eval validate {
	foreach x {alnum alpha ascii boolean digit double integer 
			lower print upper wordchar xdigit} {
		interp alias {} [namespace current]::$x {} string is $x
	}
}

proc valid {metavar val} {
	upvar #0 $metavar meta
	if {![info exists meta(type)]} { return 1 }
	foreach x $meta(type) {
		set cmd [info commands validate::$x]
		if {$cmd!="" && ![$cmd $val]} {return 0}
	}
	return 1
}

proc ref {obj} { return ::objects::[string map {: ::} $obj] }
proc obj {ref} {join [lrange [split [string map {:: "\0"} $ref] "\0"] 2 end] :}

proc get {obj key {default ""}} {
	set var [ref $obj]($key)
	if {[info exists $var]} { set $var } else { set default}
}

proc pref {name} { return ::modules::[join $name ::] }

proc meta2var {x} {
	return [namespace qualifiers [namespace qualifiers $x]]::[namespace tail $x]
}

proc var2meta {x} {
	return [namespace qualifiers $x]::meta::[namespace tail $x]
}

proc method {where obj name args} {
	set ns [uplevel 1 {namespace current}]
	if {$args==""} {
		if {[info commands ${ns}::${name}]!="${ns}::${name}"} {
			Event Log error "$name is not found in $ns"
			return
		}
	} else { eval [list interp alias {} ${ns}::${name} {}] $args }
	namespace eval $ns [list namespace export $name]
	namespace eval [ref $obj]::meta::menu [list namespace import ${ns}::$name]
}

proc select {{roots {}} {expr {}}} { 
	if {$roots==""} {
		set roots [namespace children ::objects]
	} else { set roots [map x $roots {ref $x}] }
	_select $roots $expr 
}

proc _select {roots expr} {
	set result [list]
	foreach root $roots {
		foreach x [info vars ${root}::*] {
			if {![array exists $x]} continue
			if {$expr!=""} {
				namespace eval test {}
				foreach {key val} [array get $x] { 
					set test::$key $val
				}
				set res [namespace eval test [list expr $expr]]
				namespace delete test
				if {!$res} continue
			}
			lappend result [obj $x]
		}
		foreach x [namespace children $root] {
			switch [namespace tail $x] meta continue
			set result [concat $result [_select $x $expr]]
		}
	}
	set result
}

proc changes? {var meta val} {
	if {[info exists $var]} {return  [expr {[set $var]!=$val} ]}
	if {[info exists ${meta}(default)]} {
		upvar #0 ${meta}(default) default
	} else { set default "" }
	expr {$default!=$val} 
}

# Contact command was extended to accept contacts of different messaging 
# systems. Old-style ICQ command can be distingushed by 1st argument, which is 
# numerical UIN in case of ICQ, and textual contact type otherwise.
proc Contact {args} {
	set type [expr {[string is digit [lindex $args 0]]?"old":"new"}]
	eval Contact:$type $args
}

proc Contact:old {uin alias groups {properties {}}} {
	Setter Contact:ICQ:$uin $properties [list Alias $alias Groups $groups]
}

proc Contact:new {type id properties} { new Contact:$type:$id $properties }

proc Group {gid alias {groups {}} {properties {}}} { 
	Setter Group:common:$gid $properties [list Alias $alias Groups $groups]
}

proc Setter {id plist list} {
	foreach item $plist {
		foreach {key val} [split $item =] break
		lappend list $key $val
	}
	new $id $list
}

proc new {id list} {
	set ref [ref $id]
	set noevent [info exists $ref]
	if {[catch { array set $ref $list } reason]} {
		Event Log error "Can no create object $id: $reason"
	} elseif {!$noevent} { Event New:$id $id }
}

proc nil {} { format c 0 }
proc nil? {obj} { string equal $obj [format c 0] }

# This is quite independant routine used for unified saving changes to file
proc Save {file key newval} {
	set changed 0
	set attr 0600
	if {[catch { set fd2 [open ${file}.new w] } reason] } { 
		Event Log error "Can not create ${file}.new: $reason"
		return 
	}
	if {[file exists $file]} {
		if {[catch {
			if {$::tcl_platform(platform)=={unix}} {
				set attr [file attributes $file -permissions]
			}	
			set fd1 [open $file r]
			while {[gets $fd1 ln]!=-1} {
				if {[regexp $key $ln]} {
					if {![nil? $newval]} {puts $fd2 $newval}
					set changed 1
				} else { puts $fd2 $ln }
			}
			close $fd1
		} reason]} { 
			Event Log error "Error when modifying $file: $reason"
			return
		}
	}
	if {[catch {
		if {!$changed && ![nil? $newval]} {puts $fd2 $newval }
		close $fd2
		file rename -force ${file}.new $file
		if {$::tcl_platform(platform)=={unix}} {
			file attributes $file -permissions $attr
		}
	} reason]} { Event Log error "Can't update $file: $reason"}	
}


proc rc_unix {} { list ~/.alicq/alicqrc }

proc rc_windows {} {
	set rclist [list]
	set mydir [file join [pwd] [file dirname [info script]]]
	if {[info exists ::env(USERPROFILE)]} {
		lappend rclist [file join $::env(USERPROFILE) alicq alicq.cfg]
	}
	lappend rclist [file join $mydir alicq.cfg]
	set rclist
}

proc rc_fallback {} { 
	list [file join [file dirname [info script]] alilcq.cfg]
}

# Command line parameters handling code '''1
namespace eval cmdline {
	variable -version "display Alicq version and exit"
	proc -version {} {
		puts "Alicq [package present alicq]"
		Event exit
	}
	
	variable -config "specify startup file to use"
	proc -config {filename} {
		set [namespace parent]::rcfile $filename
	}
	
	proc -info {{?module? ""}} {
		hook ConfigLoaded [list ::showmetainfo ${?module?}] .1
		hook ConfigLoaded {Event exit} .11
	}

	variable -help "display this help"
	proc -help {} {
		set txt "Alicq options:\n"
		foreach x [info procs [namespace current]::-*] {
			set call "[namespace tail $x] [info args $x]"
			set comm [expr {[info exists $x]?[set $x]:"" }]
			append txt [format "\t%-25s%s\n" $call $comm]
		}
		puts $txt
		Event exit
	}
}

# Generic handler of command-line options
proc cmdline {list} {
	while {[llength $list]} {
		set opt [lindex $list 0]
		set cmd ::cmdline::$opt
		if {[info procs $cmd]==$cmd} {
			set len [llength [info args $cmd]]
			catch { eval [list $cmd] [lrange $list 1 $len] }
			set list [lrange $list [expr $len+1] end]
		} elseif {[string match -* $opt]} {
			puts "Option $opt is unknown, skipping"
			set list [lrange $list 1 end]
		}
	}
}
#```1

proc option_info {args} {
	hook ConfigLoaded [list showmetainfo $args] .1
	hook ConfigLoaded Finis .11
}

proc option_xmlinfo {args} {
	hook ConfigLoaded [list xmlmetainfo $args] .1
	hook ConfigLoaded Finis .11
}

proc outname {name} { lrange [string map {:: " "} [meta2var $name]] 2 end }
proc showmetainfo {{names {}}} {
	if {![llength $names]} { 
		set names [namespace children ::modules]
	} else { set names [map x $names { set _ ::modules::$x }] }
	foreach ns $names {
		set meta [lsort [metainfo:module $ns]]
		if {![llength $meta]} continue
		puts "Module [namespace tail $ns]"
		set prop [list]
		foreach x $meta {
			if {![array exists $x]} {
				puts [format {%15s: %s} [outname $x] [set $x]]
			} else {lappend prop $x}
		}
		if {[llength $prop]} { puts "     Parameters:"}
		foreach x $prop {
			if {[info exists ${x}(description)]} { 
				set d [set ${x}(description)]
			} else { set d "(no description)" }
			puts [format {%20s - %s} [outname $x] $d]
			foreach field {type default} {
				if {[info exists ${x}($field)]} {
					puts [format {%30s: %s} $field [set ${x}($field)]]
				}
			}
			set value (undefined)
			catch {set value [set [meta2var $x]]} 
			puts [format {%30s: %s} current $value]
		}
		puts ""
	}
}

proc safexml {str} {
	string map [list "<" "&lt;" ">" "&gt;"] $str
}

proc xmlmetainfo {{names {}}} {
	if {![llength $names]} { 
		set names [namespace children ::modules]
	} else { set names [map x $names { set _ ::modules::$x }] }
	puts "<?xml version=\"1.0\"?>\n<metainformation>"
	foreach ns $names {
		set meta [lsort [metainfo:module $ns]]
		if {![llength $meta]} continue
		puts "\t<module name='[namespace tail $ns]'>"
		set prop [list]
		foreach x $meta {
			set name [outname $x]
			if {![array exists $x]} {
				puts "\t\t<$name>[safexml [set $x]]</$name>"
			} else {
				puts "\t\t<parameter name='$name'>"
				foreach {field val} [array get $x] {
					puts "\t\t\t<$field>[safexml $val]</$field>"
				}
				puts "\t\t</parameter>"
			}
		}
		puts "\t</module>"
	}
	puts "</metainformation>"
}

# Weird way to chech existance of namespace in pre-8.4 tcl
if {[package vsatisfies [package present Tcl] 8.4]} {
	proc nsexists {ns} { namespace exists $ns }
} else {
	proc nsexists {ns} { expr [catch { namespace parent $ns }]^1 }
}

proc LoadConfig {} {
	variable rcfile
	set base [file dirname $rcfile]
	set pwd [pwd]
	if {[catch [list cd $base] v]} {
		puts "Can not cd into $base: $v"
		return 0
	} else { set name [file tail $rcfile] }
	rename unknown unknown.config
	proc unknown {args} {
		set res ""
		if {[catch { set res [eval unknown.config $args] } v]} {
			set var [module::space [lindex $args 0]]::[join [lrange $args 1 end-1] ::]
			set metavar [var2meta $var]
			if {[info exists $metavar]} {
				set val [lindex $args end]
				if {[valid $metavar $val]} {
					set $var $val
				} else { 
					Event Log error "[lrange $args 0 end-1] is not valid [set ${metavar}(type)]"
				}
			} else {
				# Check object namespace
				set uid [join [lrange $args 0 end-1] :]
				set ns [namespace qualifiers [ref $uid]]
				if {[nsexists $ns] && $ns!="::objects"} {
					new $uid [lindex $args end]
				} else { Event Log warning $v }
			}	
		} 
		set res
	}
	set res [catch {uplevel #0 [list source $name]} v]
	rename unknown {}
	rename unknown.config unknown
	if {$res} {
		Event Log error "Error while loading startup file $name: $v"
		return 0
	} else {
		Event ConfigLoaded
		return 1
	}
}

proc configuration-object {class name list} {
	new $class:$name $list
}

# Higher-level module loading commands accessible from startup file
proc module {list} { foreach x $list { _module $x } }
interp alias {} Module {} module

proc _module {name} { 
	set name [file root [file tail $name]]
	if {[catch { set res [module::load $name] } reason]} {
		Event Log error "$reason\n$::errorInfo"
		return
	}
}

# Procedures for work with metainformation

# Generic metainfo wrapper - interface for external modules 
proc metainfo {sub args} {
	set cmd metainfo:$sub
	if {[llength [info commands $cmd]]} {
		eval [list $cmd] $args
	} else { return -code error "Invalid metainfo subcommand $sub" }
}

# Get metainformation about all modules or given module's namespace
# configurable parameters
proc metainfo:module {{root ::modules}} { module::_meta $root }

# Get list of variables containing metainformation on object, object namespace,
# or single object field
proc metainfo:object {obj {field ""}} {
	set result [list]
	for {set ns [ref $obj]} {$ns!=""} {set ns [namespace qualifiers $ns]} {
		if {$field!=""} {
			if {[info exists ${ns}::meta::${field}]} {
				return ${ns}::meta::${field}
			} else continue
		}
		foreach x [info vars ${ns}::meta::*] {
			if {[lsearch $result *::[namespace tail $x]]==-1} {
				lappend result $x
			}
		}
	}
	set result
}

proc resource {class names} {
	foreach x $names {
		if {[catch { resource::load $class $x } reason]} {
			Event Log {error resource} "$class $x loading error: $reason"
		} else { Event Log {info resource} "$class $x loaded" }
	}
}


# If sender is not in our contact list, add him into 'others' group.
handler Incoming AddToOthers {type uid args} {
	if {[info exists [ref $uid](Groups)]} return
	new $uid {}
} 0.05

# If contact has no alias when created, assign default alias.
# Protocol modules can handle this event with priority less than 0.1 
# to request alias from server (if possible).
handler {New:Contact:* New:Group:common:*} DefaultAlias  {uid} {
	upvar 0 [ref $uid](Alias) alias
	if {[info exists alias]} return
	puts "Set default alias"
	set alias [lindex [split $uid :] end]
} 0.1

handler {New:Contact:* New:Group:common:*} DefaultGroup  {uid} {
	upvar 0 [ref $uid](Groups) groups
	if {[info exists groups]} {
		foreach x $groups {
			if {![info exists [ref Group:common:$x]]} {
				new Group:common:$x [list Alias $x]
			}
		}
		return
	} else {
		if {[string match Contact:* $uid]} {
			set groups other
		} else { set groups "" }
	}
} 0.1

handler Error Error {id {description "Unknown error"}} {
	Event Log error "ERROR $id \($description\)" 
	list $id $description
}

handler exit Finis {args} { 
	exit
} 0.999

proc InitRC {} {
	variable rcfile
	# If no startup file given as command-line option
	# search by platform-dependent list of candidates
	if {![info exists rcfile]} {
		set cmd rc_$::tcl_platform(platform)
		if {[info commands $cmd]==$cmd} {
			set candidates [$cmd]
		} else { set candidates [rc_fallback] }

		# No candidates found - something is wrong
		if {![llength $candidates]} {
			Error internal "Do not know where config file should be"
			Finis
		}

		# Trying to find existing file
		foreach name $candidates { 
			if {[file exists $name]} { set rcfile $name; break } 
		}

		# If no existing file found, take first candidate as a target
		# for new confug
		set rcfile [lindex $candidates 0]
	}
}

proc alicq_init {} {
	set mydir [file join [pwd] [file dirname [info script]]]
	# Detect available data paths
	set paths [list]
	foreach x [list [file dirname [rcfile]] [file join $mydir data]\
			[file join $mydir .. share alicq]] {
		if {[file isdirectory $x]} { lappend paths $x }
	}
	resource::path $paths
	# Internatinalization
	package require msgcat
	namespace import ::msgcat::*
	foreach x $paths {
		if {[msgcat::mcload [file join $x messages]]} break 
	}
	module::path [map x $paths { file join $x modules } ]
	set autoload [file join [file dirname [rcfile]] autoload]
	module::autoload $autoload
	if {[file isdirectory $autoload]} {
		Event Log info "Automatical configuration ability detected"
		module::load config
	}
	variable rcfile
	if {[file exists $rcfile]} {
		LoadConfig
	} elseif {![file isdirectory $autoload]} {
		# In case no config found, and no autoload, 
		# enable default modules
		set base [file dirname $rcfile]
		if { ![file isdirectory base] } {
			if {[catch { file mkdir $base } r ]} {
				Event Log error "Can not create base dir: $r"
			} else {
				foreach x $::default_modules {
					if {![catch { module::load $x }]} {
						module::enable $x
					}
				}
			}
		}
	}
}

# If no or empty parameter is given - return name of startup file used,
# otherwise - treat parameter as name of the file which should be located in
# startup file base directory
proc rcfile {{name ""}} {
	variable rcfile
	if {$name==""} { return $rcfile }
	file join [file dirname $rcfile] $name
}

# Parse command-line options
if {[info exists argv]} { cmdline $argv }
# Check for default rc file
InitRC
# Default loggind
modules::log::init [rcfile].log


