
package provide resource 1.0

namespace eval resource {
	variable path [list]
}

proc resource::path {args} {
	variable path
	if {[llength $args]} {
		foreach x $args { set path [concat $path $x] }
	}
	foreach x [classes] { rehash $x }
}

# Get list of all registered classes
proc resource::classes {} {
	variable class
	set res [list]
	foreach x [array names class *,script] {
		lappend res [string range $x 0 end-7]
	}
	set res
}

proc resource::class {name script args} {
	variable class
	if {[info exists class($name,command)]} {
		return -code error "Resource class $name already registered"
	}
	set class($name,script) $script
	if {[info commands option]!=""} {
		set class($name,theme) [option get .\
			"[string tolower $name]Theme" Theme]
	} else { set class($name,theme) "" }
	if {[llength $args]} { set class($name,mask) $args }
	rehash $name
}

proc resource::unregister {name} {
	variable class
	if {[info exists class($name,command)]} {
		return -code error "Resource class $name already registered"
	}
	foreach item [array get class($name,loaded,*)] {
		set item [lindex [split $item ,] end]
		catch { eval $class($name,script) unload $item }
	}	
	array unset class $name,*
}

# Internal procedure to update resource hash.
proc resource::rehash {name} {
	variable class
	if {![info exists class($name,script)]} {
		return -code error "Resource class $name is not registered"
	}
	variable path
	variable cache
	array unset cache $name,*
	if {[info exists class($name,mask)]} {
		 set mask $class($name,mask)
	} else { set mask * }
	if {$class($name,theme)!=""} {
		foreach dir $path {
			_cacheupdate $name [file join $dir $name\
				$class($name,theme)] $mask
		}
	}	
	foreach dir $path {
		_cacheupdate $name [file join $dir $name] $mask
	}
}

proc resource::_cacheupdate {class dir masks} {
	variable cache
	foreach mask $masks {
		foreach x [glob -nocomplain [file join $dir $mask]] {
			set name [file tail [file root $x]]
			if {[info exists cache($class,$name)]} continue
			set cache($class,$name) $x
		}
	}
}

# Get list of available themes
proc resource::themes {name} {
	variable path
	set res [list]
	foreach dir $path {
		foreach x [glob -nocomplain -type d [file join $dir $name *]] {
			lappend res [file tail $x]
		}
	}
	set res
}

# Cuery qurrent theme or set new one
proc resource::theme {name args} {
	variable class
	if {[llength $args]==1} {
		set class($name,theme) [lindex $args 0]
		rehash $name
		reload $name
	}
	set class($name,theme)
}

# Load required resources
proc resource::load {name item} {
	variable class
	if {![info exists class($name,script)]} {
		return -code error "Resource class $name is not registered"
	}
	variable cache
	if {[info exists class($name,loaded,$item)]} return
	if {[info exists cache($name,$item)]} {
		if {[catch { eval $class($name,script) load\
			$item $cache($name,$item) } reason]} {
			return -code error "Resource loader error: $reason"
		} else { set class($name,loaded,$item) $cache($name,$item) }
	} else {
		return -code error "$name $item is not found in search path"
	}
}

proc resource::unload {name item} {
	variable class
	if {![info exists class($name,script)]} {
		return -code error "Resource class $name is not registered"
	}
	if {[info exists class($name,loaded,$item)]} {
		if {[catch { eval $class($name,script) unload\
				$item} reason]} {
			return -code error "Resource unloader error: $reason"	
		} else { unset class($name,loaded,$item) }
	} else { return -code error "$name $item is not loaded" }
}

# Reload all previously loaded resources of given class if their origin is
# changed
proc resource::reload {name} {
	variable class
	variable cache
	foreach {x from} [array get class $name,loaded,*] {
		set x [lindex [split $x ,] end]
		if {[info exists cache($name,$x)] &&
		    $cache($name,$x)!=$from } { 
		    	unload $name $x 
			load $name $x
		}
	}
}

proc resource::origin {name item} {
	variable class
	if {![info exists class($name,loaded,$item)]} {
		return -code error "$name $item is not loaded"
	}
	set class($name,loaded,$item)
}

