#  Alicq ICQ client: sound notification module 
#  Copyright (C) Ihar Viarheichyk 2001

#  This module is a part of Alicq ICQ client.

package require resource

variable players {auplay esdplay artsplay bplay play aplay}

namespace eval meta {
	set author "Ihar Viarheichyk <iverg@mail.ru>"
	set description "Basic implementation of sound notifications."
	array set player {description "External player" save change}
}

# Search by the list of well-known unix players in the path
proc FindExternalPlayer {} {
	variable player
	variable players
	if {[set x [External $players]]!=""} {
		set player $x
		Event Log info "Found external player $player"
		return 1
	} else { return 0 }
}

# For external player no loading is required, just invoke external command
# with resource file name
proc ExternalPlayer {cmd args} {
	variable player
	if {$cmd=="play" && $player!=""} {
		set name [lindex $args 0]
		if {[catch {exec $player [resource::origin sounds $name]} r]} {
			Event Log {error sound} $r
		}
	}	
}

proc InternalPlayer {cmd args} {
	switch -- $cmd {
		load {  foreach {name file} $args break
			snack::sound sound:$name -file $file
		}
		play { sound:[lindex $args 0] play -blocking no }
		init { append meta::player(description) "\n(snack is in use now, external player will be available after Alicq reload)"
		}
	}
}

proc init {args} {
	variable player
	if {![info exists player] || $player==""} {
		# If no external player specified, trying to load snack
		if {[catch { package require snack } reason]} {
			Event Log error "Snack loading error: $reason"
			# Snack was not loaded, try to find external player
			# automatically
			if {$::tcl_platform(platform)=="unix" &&
			    [FindExternalPlayer]} { set cmd ExternalPlayer }
		} else { set cmd InternalPlayer }
	} else { set cmd ExternalPlayer }
	if {[info exists cmd]} {
		resource::class sounds [namespace code $cmd] *.wav
		resource sounds {Message Auth}
		$cmd init
		set ns [namespace current]
		interp alias {} ${ns}::play {} ${ns}::$cmd play
		handler Incoming PlaySound {type args} {
		}
	}
}
after idle [nc init]
