# This module is a part of Alicq instant messenger package
# It provides core of basic graphical interface of Alicq: tree window
#
# Author: Ihar Viarheichyk

# Contact list interface - it .tr widget exists, another contact list
# module is loaded
if {[winfo exists .tr]} {
	return -code "Contact list is already initialized, skipping"
}

package require BWidget
package require resource

#module {hideoffline hideempty}
variable alias unknown
array set hidden [list]

event add <<Hide-Offline>> <Control-o>
event add <<Hide-Empty>> <Control-i>
event add <<ContextCurrent>> <3> <Control-1>
event add <<ContextSelected>> <F9>
event add <<Item-Activate>> <Double-1> <Return>
event add <<QuickSearch>> <Key-slash> <Control-Key-f>
event add <<SearchCancel>> <FocusOut> <Key-Escape>

option add *quick.Checkbutton.selectColor "" widgetDefault
option add *quick.Checkbutton.indicatorOn 0
#option add *status.compound [expr {([tk windowingsystem]=="aqua")?"left":"none"}]
option add *status.compound left
option add *quickPanel true widgetDefault
option add *rosterContactColor black widgetDefault
option add *localContactColor blue widgetDefault

namespace eval meta {
	variable author "Ihar Viarheichyk <iverg@mail.ru>"
	variable name "Contact list"
	variable description "Basic group and contact list implementation"
	variable icon img:group
	array set grouping {
		type variant default Plain property Global:UI:Tree|Grouping
		description "Grouping rule" save exit
	}

	foreach x {Contact Group} {
		method menu $x [list 0.98 Rename] [namespace parent]::Rename
		method menu $x [list 0.99 Delete] [namespace parent]::Delete
	}

	namespace eval menu {
		variable [list Interface Grouping] grouping\
			 [list Interface "Hide offline contacts"] hide::offline\
			 [list Interface "Hide empty groups"] hide::empty\
			 {0.099 Search {In Contact List}} quick_search
	}

	proc unloadable {} { return 0}
}

set meta::grouping(valuescript) [nc Available Grouping]

#set grouping Plain

namespace eval hide {
	namespace eval meta {
		array set empty {
			type boolean default 0 property Global:UI:Tree|HideEmpty
			description "Hide empty groups" save exit
			
		}
		array set offline {
			type boolean default 0
			property Global:UI:Tree|HideOffline
			description "Hide offline contacts" save exit
		}
	}
}

# Visibility filters
namespace eval filter {
	# Contacts with pending messages are always visible
	proc _pending {uid} {
		if {[info exists [ref $uid](Pending)]} {
			return -code break
		} else { return 1}
	}
}

namespace eval [ref Group:common]::meta {
	array set Open	{type boolean default 0 save exit} }
namespace eval [ref Group:status] {}

proc MyDraw {path} { 
	_draw_tree_old $path
	#set im [.tr.c create image 0 0 -anchor nw -image [Bitmap::get /usr/share/pixmaps/Snail.xpm]]
	#.tr.c lower $im
}

proc DeltaY {} {
	set max 0
	foreach x [image names] {
		set h [image height $x]
		if {[string match img:* $x] && $h>$max} {set max $h}
	}
	incr max 2
}

# Main tree window
proc MainWindow {} {
	wm group . .
	wm command . [file join [pwd] $::argv0]
	# Handle geometry command-line option
	if {![info exists ::geometry] &&
	    [set gm [option get . geometry Geometry]]!=""} {wm geometry . $gm }
	# Handle state of main window on startup: normal or iconic	 
	if {[set state [option get . state State]]!=""} { wm state . $state }
	set searchVar [namespace current]::quick_search
	# Create scrolled tree widget
	frame .quick
	if {[option get . quickPanel Boolean]=="true"} {
		grid .quick -columnspan 2 -sticky we
	}
	grid columnconfigure .quick 8 -weight 1
	grid [checkbutton .quick.offline -image img:offline\
		-variable [namespace current]::hide::offline]\
		-row 0 -column 1 -padx 2 -ipadx 2 -ipady 2 -sticky w
	ui::tooltip::action .quick.offline\
		[mc $hide::meta::offline(description)] <<Hide-Offline>>
	grid [checkbutton .quick.empty -image img:group\
		-variable [namespace current]::hide::empty]\
		-row 0 -column 2 -padx 2 -ipadx 2 -ipady 2 -sticky w
	ui::tooltip::action .quick.empty\
		[mc $hide::meta::empty(description)] <<Hide-Empty>>
	grid [checkbutton .quick.search -text [mc Search] -variable $searchVar]\
		-row 0 -column 10 -padx 2 -ipadx 2 -ipady 2 -sticky e
	ui::tooltip::action .quick.search\
		[mc "Search in contact list"] <<QuickSearch>>
	# Add command to main menu	
	namespace eval meta::menu { 
		proc {Search "In Contact List"} {} { .quick.search invoke }
	}	
	trace variable $searchVar w [nc QuickSearch .qsearch .tr.c]
	set sw [ScrolledWindow .sw -relief flat -borderwidth 0]
	set tree [Tree .tr -deltay [DeltaY]\
		-dragenabled yes -dropenabled yes -dropcmd [nc DropCmd]\
		-opencmd [nc OpenCmd 1] -closecmd [nc OpenCmd 0]]
	$sw setwidget $tree
	foreach id {Text Image} {
		$tree bind${id} <Double-1> [nc DoubleClick]
		$tree bind${id} <<ContextCurrent>> [nc PopupMenu %X %Y]
		$tree bind${id} <Enter> [nc ShowTip]
		$tree bind${id} <Leave> [nc CloseTip]
	}
	foreach x {roster local} { 
		variable _${x} [option get .tr ${x}ContactColor ContactColor] 
	}

	bind .tr.c <<ContextSelected>> [nc PopupSelected %X %Y]
	if {[catch { [tk windowingsystem] }]=="aqua"} {
		menu .main -type menubar -tearoff no
		. configure -menu .main
		.main add cascade -menu .main.apple
		.main add cascade -menu .main.context
		bind .tr <<TreeSelect>> [nc NodeSelected %W]
	} else {
		set menu1 [menubutton .main -text [mc "Menu"]\
			-direction above -menu .main.apple]
		ui::tooltip::action .main [mc "Invoke main menu"]
		grid $menu1 -column 1 -row 5 -sticky nes -padx 2
	}
	menu .main.apple -type normal -tearoff no
	set menu2 [menubutton .status -text [mc offline]\
		-menu .status.m -image img:offline]
	set search [FastSearch .qsearch .tr $searchVar]
	grid $sw -columnspan 2 -row 3 -sticky news
	grid $menu2 -column 0 -row 5 -sticky nws -padx 2 
	grid columnconfigure . 1 -weight 1
	grid rowconfigure . 3 -weight 1

	bind .tr.c <<QuickSearch>> [list set $searchVar 1]
	bind .tr.c <<Popup>> { puts "Key %K" }

	# Small hack to allow mouse wheel scrolling in tree widget
	bind .tr.c <4> ".tr.c yview scroll -2 units"
	bind .tr.c <5> ".tr.c yview scroll 2 units"
	
	if {[wm protocol . WM_DELETE_WINDOW]==""} {
		wm protocol . WM_DELETE_WINDOW {Event exit}
	}
	bind . <Return> [namespace code  {DoubleClick [.tr selection get]}]
	bind [winfo class .] <FocusIn> { focus .tr.c }
}

proc ShowTip {node} { ui::tooltip::show [nc MakeTip [UidByNode $node]] }
proc CloseTip {args} { ui::tooltip::close }

proc MakeTip {uid} {
	upvar #0 [ref $uid] info
	foreach {class type id} [split $uid :] break
	set t "$type id: $id"
	if {[info exists info(Client)]} {
		foreach {ver client unicode} $info(Client) break
		append t "\n$client, ICQv${ver}\n"\
			[expr ${unicode}?"Unicode":"Encoding: [get $uid encoding [set [pref {icq encoding}]]]"]
	}
	if {[info exists info(client:name)]} {
		append t "\n$info(client:name)" 
	}
	if {[info exists info(Status)]} {
		append t "\n[mc Status]: [mc $info(Status)]"
	}
	if {[info exists info(description)]} {
		append t "($info(description))"
	}
	if {[info exists info(IP)]} { append t "\nIP: $info(IP)" }
	if {[info exists info(dc)] && [info exists info(LocalIP)]} {
		append t "\nDC IP: $info(LocalIP)"
	}
	set t
}		

proc QuickSearch {name main var args} {
	upvar 1 $var mode
	set current [winfo viewable $name]
	if {$current==$mode} return 
	variable last_pos
	if {$mode} {
		grid $name -column 0 -columnspan 2 -row 4 -sticky we
		set last_pos 0
		#$name selection range 0 end
		focus $name
	} else { 
		variable last_search
		grid forget $name 
		set val [$name get]
		if {$val!=""} {
			if {[info exists last_search]} {
				set pos [lsearch $last_search $val]
				if {$pos!=-1} { 
					set last_search [lreplace\
						$last_search $pos $pos]
				}
			}
			lappend last_search $val
		}
		if {[info exists last_pos]} { unset last_pos }
		$name delete 0 end
		focus $main
	}
}

proc FastSearch {top tree var} {
	entry $top -validate key -vcmd [nc DoSearch $top %P $tree]
	bind $top <<SearchCancel>> [list set $var 0]
	bind $top <Up> [nc SearchHistory $top]
	set top
}

proc SearchHistory {name} {
	variable last_search
	variable last_pos
	if {![info exists last_search]} return
	set cur [$name get]
	set next [lindex $last_search end-$last_pos]
	if {$next==""} { set last_pos 0 } else { incr last_pos }
	$name delete 0 end
	$name insert end $next
}

proc DoSearch {entry val tree} {
	if {$val==""} { return 1}
	set found [list]
	set idx [string length $val]
	foreach uid [select Contact] {
		set alias [get $uid Alias]
		set uin [lindex [split $uid :] end]
		if {[string equal -nocase -length $idx $val $alias]} {
			lappend found $uid
		} elseif {[string equal -length $idx $val $uin]} {
			lappend found $uid
		} 	
	}
	if {[llength $found]} { return [ShowFound $tree $found] }
	return 0
}

proc ShowFound {tree items} {
	variable hidden
	variable mapping
	foreach uid $items {
		if {[info exists hidden($uid)]} continue
		set r [lindex $mapping($uid) 0]
		if {[$tree visible $r]} break
	}
	if {![info exists r]} { return 0 }
	if {![$tree visible $r]} { RecOpen $tree $r }
	$tree see $r
	$tree selection set $r
	return 1
}

proc RecOpen {tree node} {
	if {$node=="root"} return
	$tree itemconfigure $node -open 1
	RecOpen $tree [NodeParent $node]
}

proc OpenCmd {val node} { set [ref [UidByNode $node]](Open) $val }

proc DoubleClick {node} {
	if {[set uid [UidByNode $node]]=={}} return
	# If  node is group: open/close
	if [string match Group:common:* $uid] {
		.tr itemconfigure $node -open\
			[set [ref $uid](Open) [expr {[get $uid Open 0]^1}]]
	} else { Event $uid|Send:text $uid|Send:text }
}

proc CreateMenus {} {
	menu .status.m -type normal -tearoff no
	set noimage [string equal [catch { [tk windowingsystem]}] "aqua"]
	foreach status {online away dnd occ ffc na invisible offline} {
		.status.m add command -label [mc $status]\
			-command "Event SetStatus $status"
		if {!$noimage} { .status.m entryconfigure last -image img:$status }
		set [ref Group:status:$status](Alias) [mc $status]
		nodecache Group:status:$status
	}
}

# --- Handle in-tree item renaming
proc Rename {uid} {
	variable CurrentNode
	.tr edit $CurrentNode [get $uid Alias] [nc ItemModify $uid]
}

proc ItemModify {uid text} {
	set [ref $uid](Alias) $text
	return 1
}

# --- Handle moving items between groups
proc DropCmd {dest src type op datatype data} {
	variable counters
	# Don't accept unknown types and don'w allow loops
	if {$type=="widget" && $dest==".tr" ||
	    [lindex $type 0]=="position" && [lindex $type 1]=="root"} {
		set uid [UidByNode $data]
		set parent [UidByNode [NodeParent $data]]
		if {[string match Group:common:* $parent]} {
			counter $uid $parent -1
			set parent [lindex [split $parent :] end]
			set [ref $uid](Groups) [grep x [get $uid Groups {}]\
				{$x!=$parent}]
		}
		return
	}
	if {![string match node* $type]||\
	     [string match ${data}* [lindex $type 1]]} return
	set parent_n [UidByNode [lindex $type 1]]
	switch -glob -- $parent_n {
		Group:common:* {
			set uid [UidByNode $data]
			if {$uid==$parent_n} return
			set dest [lindex [split $parent_n :] end]
			set parent_o [UidByNode [NodeParent $data]]
			set old [lindex [split $parent_o :] end]
			# Set filter expression depending of drop type
			if {$op!="copy"} {
				set exp {$x!=$old}
				counter $uid $parent_o -1
			} else { set exp 1 }
			counter $uid $parent_n 1
			append exp {&& $x!=$dest}
			set [ref $uid](Groups) [concat\
				[grep x [get $uid Groups {}] $exp] $dest]
		}
	}
}

# --- Handle deletion of an item
proc Delete {id} {
	variable CurrentNode
	variable mapping
	set uid [UidByNode $CurrentNode]
	if {[set pos [lsearch -exact $mapping($uid) $CurrentNode]]>=0} {
		set lst [lreplace $mapping($uid) $pos $pos]
		if {![llength $lst]} { 
			set name [get $uid Alias]
			if {[tk_messageBox -type yesno -title "Removing $name"\
				-message "Remove $name from contact list?"]
					!="yes"} return
			hide $uid 1		
			unset mapping($uid)
			unset [ref $uid](Groups)
			#after idle [list unset [ref $uid]]
			unset [ref $uid]
		} else {
			set parent [UidByNode [NodeParent $CurrentNode]]
			counter $uid $parent -1
			set parent [lindex [split $parent :] end]
			set [ref $uid](Groups) [grep x [get $uid Groups {}] {
				$x!=$parent
			}]
		}
	}
}

proc counter {uid parent incr} {
	variable counters
	if {![info exists counters($parent,total)]} return
	incr counters($parent,total) $incr
	if {[string match Contact:* $uid] && 
	    [get $uid Status offline]!="offline"} {
	    	incr counters($parent,online) $incr
	}
	NeedUpdate $parent text
}

# --- Context menu handling
proc PopupSelected {x y} {
	set sel [.tr selection get]
	if {[llength $sel]} { PopupMenu $x $y [lindex $sel 0] }
}

# Display node context menu as dropdown of application menu
catch {
if {[tk windowingsystem]=="aqua"} {
	proc NodeSelected {tree} {
		set nodes [$tree selection get]
		if {[llength $nodes]!=1} return
		set node [lindex $nodes 0]
		MakeMenu .main.context [UidByNode $node]
		.main entryconfigure last -label [$tree itemcget $node -text]
		variable CurrentNode
		set CurrentNode $node
	}
}
}

proc PopupMenu {x y node} {
	variable CurrentNode
	MakeMenu .context [UidByNode $node]
	if {![winfo exists .context]||[.context index last]=="none"} return
	set CurrentNode $node
	tk_popup .context [expr $x-10] [expr $y-5]
}

proc init {} {
	resource icons {group away dnd ffc invisible
			message na occ offline online url}

	MainWindow
	CreateMenus

	bind . <<Hide-Offline>> [nc trigger hide::offline]
	bind . <<Hide-Empty>> [nc trigger hide::empty]

	trace variable hide::empty w [nc HideEmpty]
	trace variable hide::offline w [nc trigger_offline]

	trace variable [ref Me](Status) w [nc UpdateStatus]
	trace variable [ref Me](Alias) w [nc WindowTitle]
	UpdateStatus

	# If no menu created by external modul add at least exit entry
	if {[.main.apple index last]=="none"} { 
		.main.apple add command -label [mc Exit] -command {Event exit}
	}
	after idle [nc Populate]
}

proc Available {ns} {
	map x [info commands ${ns}::*] { namespace tail $x }
}

proc Populate {} {
	variable grouping
	handler {ModuleLoaded ModuleUnloaded} recalcMenu {module} { 
		MakeMenu .main.apple 
	} 0.99
	trace variable grouping w [nc Regroup]
	if {![info exists [ref Group:common:other]]} { Group other Other }
	
	GroupingChanged

	foreach x {contacts groups} {
		SortAlias $x [set sort::$x]
		trace variable sort::$x w [nc resort] 
	}
	MakeMenu .main.apple

	foreach x [select Group:common] { CommonMonitor $x; GroupMonitor $x }
	foreach x [select Contact] { CommonMonitor $x; ContactMonitor $x}
	hook {New:Contact:* New:Group:common:*} [nc CommonMonitor] 0.6
	hook New:Contact:* [nc ContactMonitor]
	hook New:Group:common:* [nc GroupMonitor]
}

proc hide {uid {mode 1}} {
	variable hidden
	variable mapping
	set exists [info exists hidden($uid)]
	if {![info exists mapping($uid)]} return
	if {$mode && !$exists} {
		foreach x $mapping($uid) { do_hide $x }
		set hidden($uid) $mode
	} elseif {$exists && !$mode} {
		foreach x $mapping($uid) { do_show $x $uid }
		unset hidden($uid)
	}
}

proc do_hide {node} {
	if {![.tr exists $node]} return
	.tr selection remove $node
	.tr delete $node 
	set parent [NodeParent $node]
	if {$parent=="root"} return
	set group [UidByNode $parent]
	if {[string is true $hide::empty] && ![llength [.tr nodes $parent]]} { 
	    	hide [UidByNode $parent] 
	} elseif {[string match Group:common:* $group]} {
		NeedUpdate $group text 
	}
}

proc do_show {node uid} {
	set parent [NodeParent $node]
	if {$parent!="root"} { 
		set group [UidByNode $parent]
		hide $group 0 
		NeedUpdate $group text
	}
	set opts [list]
	variable nodecache
	set off [expr [string length $uid]+1]
	foreach {key val} [array get nodecache $uid,*] {
		lappend opts [string range $key $off end] $val
	}
	set pos 0
	foreach n [.tr nodes $parent] {
		if {[comparator $n $node]>=0} break else { incr pos }
	}
	eval [list .tr insert $pos $parent $node] $opts
}

proc CommonMonitor {uid} {
	variable mapping
	variable hidden
	trace variable [ref $uid](Alias) w [nc onAlias $uid]
	trace variable [ref $uid](rid) wu [nc onRid $uid]
	trace variable [ref $uid] u [nc Hide $uid]
	set hidden($uid) 1
}

proc ContactMonitor {uid} {
	variable counters
	foreach x [get $uid Groups] { incr counters(Group:common:$x,total) }
	trace variable [ref $uid](Status) w [list after idle [nc onStatus $uid]]
	trace variable [ref $uid](Pending) w [nc DoPending $uid]
	trace variable [ref $uid](Groups) w [nc Mapping $uid]
	nodecache $uid
	Mapping $uid
	hide $uid $hide::offline
}

proc GroupMonitor {uid} {
	variable counters
	foreach x {total online} { set counters($uid,$x) 0 }
	trace variable [ref $uid](Groups) w [nc MapChildren $uid]
	nodecache $uid
	MapChildren $uid
	hide $uid $hide::empty
}

proc onRid {uid name1 name2 op} { 
	NeedUpdate $uid fill
	if {$op=="u" && [info exists [ref $uid](Groups)]} {
		trace variable [ref $uid](rid) wu [nc onRid $uid]
	}
}

proc onAlias {uid args} {
	NeedUpdate $uid text
	DoSort $uid
}

proc onStatus {uid name1 name2 op} {
	upvar 1 ${name1}($name2) status
	variable old_status
	set exists [info exists old_status($uid)]
	if {$status=="offline" && $exists} {
		unset old_status($uid)
		set incr -1
	} elseif {!$exists && $status!="offline"} { set incr 1 }
	if {$status!="offline"} { set old_status($uid) $status }
	if {[info exists incr]} {
		variable counters
		foreach x [get $uid Groups] {
			incr counters(Group:common:$x,online) $incr 
			NeedUpdate Group:common:$x text
		}
	}
	# Remap if it is required
	variable grouping
	set var Grouping::$grouping
	if {[info exists $var] && [lsearch -exact [set $var] Status]!=-1} {
		puts "Do remap on status"
		Mapping $uid 
	}
	variable pending
	set hide [expr {$hide::offline && $status=="offline" && 
	                ![info exists [ref $uid](Pending)]}]
	hide $uid $hide
	NeedUpdate $uid image
	if {!$hide && $sort::contacts=="Status"} { DoSort $uid }
}

variable needupdate [list]

proc NeedUpdate {uid value} {
	variable needupdate
	variable timer
	set val [list $uid $value]
	if {[lsearch -exact $needupdate $val]==-1} {
		lappend needupdate $val
		if {![info exists timer]} { 
			set timer [after idle [namespace current]::DoUpdate] 
		}
	}
	
}

proc DoUpdate {} {
	variable needupdate
	variable timer
	variable mapping
	variable hidden
	foreach x $needupdate {
		foreach {uid type} $x break
		foreach {key val} $type break     
		if {$val==""} { set val [node-$type $uid] }
		variable nodecache
		set nodecache($uid,-$key) $val
		if {![info exists mapping($uid)] ||
		     [info exists hidden($uid)]} continue
		foreach node $mapping($uid) {
			.tr itemconfigure $node -$key $val
		}
	}
	unset timer
	set needupdate [list]
}

proc Hide {uid ref field op} {
	if {$field==""} { MappingChanged $uid [list] }
}

proc FlashIcon {uid stage} {
	NeedUpdate $uid [list image\
		img:[expr {$stage?[get $uid Status offline]:"message"}]]
	set stage
}

# --- Flashing icon when pending messages present
proc NoPending {uid hid args} {
	unhook $hid
	FlashIcon $uid 1
	Flash less
	#Mapping $uid
	hide $uid [expr {$hide::offline && [get $uid Status offline]=="offline"}]
	trace variable [ref $uid](Pending) w [nc DoPending $uid]
	# Check if there other invisble contacts with pending messages
	delayed [nc SeeNext]
}

proc DoPending {uid args} {
	#if {![info exists mapping($uid)]} { Mapping $uid }
	See $uid
	set hid [hook Flashing [nc FlashIcon $uid]]
	Flash more
	trace vdelete [ref $uid](Pending) w [nc DoPending $uid]
	trace variable [ref $uid](Pending) u [nc NoPending $uid $hid]
}

# Find contact havind pending message and make it viewable in the tree
proc SeeNext {} {
	set uid [lindex [select Contact {[info exists Pending]}] 0]
	if {$uid!=""} { See $uid }
}

# Resisively open parent groups and make any of object nodes visible
proc See {uid} {
	variable mapping
	if {[info exists mapping($uid)]} {
		hide $uid 0
		foreach x $mapping($uid) { if {[.tr visible $x]} break }
		if {![.tr visible $x]} { RecOpen .tr $x}
		.tr see $x
	}
}

# Namespace containing different grouping policies
namespace eval Grouping {
	# No grouping (default)
	proc Plain {uid} {
		expr {[string match Contact:* $uid]?{root}:{}}
	}
	# Use grous
	variable Groups [list Groups]
	proc Groups {uid} {
		set ref [ref $uid]
		if { ![info exists ${ref}(Groups)] } { return "" }
		set g [set ${ref}(Groups)] 
		expr {[llength $g]?[map x $g {set _ Group:common:$x}]:{root}}
	}
	variable Status [list Status]
	# Group by status
	proc Status {uid} {
		switch -glob -- $uid {
			Contact:* { return Group:status:[get $uid Status offline]}
			Group:status:* {return root}
			* {return {}}
		}
	}
	variable Online [list Status]
	# Divide into two groups: online and offline contacts
	proc Online {uid} {
		switch -glob -- $uid {
			Contact:* {
			   return Group:status:[expr {([get $uid Status offline]==\
			   		{offline})?{offline}:{online}}]
			}
			Group:status:online {return root}
			Group:status:offline {return root}
			* {return {}}
		}
	}
}

proc Regroup {args} {
	GroupingChanged
	Remap
}

proc GroupingChanged {} {
	variable grouping
	interp alias {} grouper {} [namespace current]::Grouping::$grouping
	# Hide lines in plain mode and show if necessary in others
	if {$grouping=="Plain"} {
		.tr configure -showlines 0 -deltax 0
	} else {
		.tr configure -showlines [option get .tr showlines D]\
		 	      -deltax [option get .tr deltax D]
	}
}

proc Remap {args} { foreach x [select {Group Contact}] { Mapping $x } }

proc HideEmpty {args} {
	variable mapping
	variable hidden
	foreach {key val} [array get mapping Group:*] {
		set exists [info exists hidden($key)]
		if {$exists && !$hide::empty} {
			hide $key 0
		} elseif {$hide::empty && !$exists} {
			set n [lindex $val 0]
			if {$n!=""  && ![llength [.tr nodes $n]]} { hide $key }
		}
	}
}

proc nodecache {uid} {
	variable all_node_options
	variable nodecache
	foreach x $all_node_options { set nodecache($uid,$x) [node$x $uid] }
}

proc UidByNode {node} { lindex [split $node "\0"] end }

proc NodeParent {node} {string range $node 0 [expr [string last "\0" $node]-1]}

proc node-text {uid} { 
	set text [get $uid Alias]
	if {[string match Group:common:* $uid]} {
		variable counters
		append text " - $counters($uid,online)/$counters($uid,total)"
	}
	set text
}

proc node-image {uid} {
	if {[string match Contact:* $uid]} { 
		set img [get $uid Status offline] 
	} else { set img group }
	#if {[nil? $img]} { set img offline }
	return img:$img
}

proc node-fill {uid} {
	upvar #0 [ref $uid](rid) rid
	set [namespace current]::[expr {[info exists rid]?"_roster":"_local"}]
}

proc node-open {uid} { 
	if {[string match Group:* $uid]} { 
		get $uid Open 0
	} else { return -code continue }
}

# Sort namespace includes all available sorting methods
namespace export UidByNode
namespace eval sort {
	namespace eval meta {
		variable contacts
		array set contacts {
			type variant default Status save exit
			description "Sorting rule for contacts"
			property Global:UI:Tree|Sort:Contacts
		}
		variable groups
		array set groups {
			type variant default Identifier save exit
			description "Sorting rule for groups"
			property Global:UI:Tree|Sort:Groups
		}
	}

	# Add to main menu (temporary)
	namespace eval [namespace parent]::meta::menu {
		variable [list Interface "Sort contacts by"] sort::contacts\
			 [list Interface "Sort groups by"] sort::groups
	}

	
	namespace export *
	# No sorting at all
	proc Unsorted {s1 s2} {return 0}
	# Sort by identifier (e.g UIN for ICQ contacts, group ID for groups)
	proc Identifier {s1 s2} { string compare $s1 $s2 }
	# Sort by alphabet (case-sensitive)
	proc Alphabet {s1 s2} {
		string compare	[get $s1 Alias] [get $s2 Alias]
	}
	# Sort by alphabet (case-insensitive)
	proc {Alphabet,case-insensitive} {s1 s2} {
		string compare -nocase [get $s1 Alias] [get $s2 Alias]
	}
	# Contact-specific sorting
	namespace eval contacts {
		namespace import [namespace parent]::*
		# Sort by status and then by alphabet
		array set weight {ffc a online b occ c dnd d away e
				na f invisible g offline z}
		proc Status {s1 s2} {
			variable weight
			set st1 [get $s1 Status offline]
			set st2 [get $s2 Status offline]
			set res [string compare $weight($st1) $weight($st2)]
			if {$res} {set res} else { Alphabet $s1 $s2 }
		}
	}
	#Group-specific sorting
	namespace eval groups {
		namespace import [namespace parent]::*

		if 0 {
		proc Unsorted {s1 s2} {
			puts "compare $s1 and $s2"
			set w1 [get $s1 weight 0.5]
			set w2 [get $s2 weight 0.5]
			expr {($w1<$w2)?-1:(($w1==$s2)?0:1)}
		}
		}
	}
}
set sort::meta::contacts(valuescript) [nc Available sort::contacts]
set sort::meta::groups(valuescript) [nc Available sort::groups]

proc comparator {n1 n2} {
	foreach x {1 2} {
		set node$x [UidByNode [set n$x]]
		set type$x [lindex [split [set node$x] :] 0]
	}
	if {[set res [string compare $type1 $type2]]} { return $res }
	sort${type1} $node1 $node2
}

proc sort {x} {
	if {![.tr exists $x]} return
	.tr reorder $x [lsort -command comparator [.tr nodes $x]]
	# workaround for nonupdating tree in Plain mode
	if {$x=={root}} {.tr insert end root _fake_; .tr delete _fake_}
}

proc resort {name args} {
	variable mapping
	upvar 1 $name val
	SortAlias [namespace tail $name] $val
	foreach {uid nodes} [array get mapping Group:*] { 
		foreach node $nodes { delayed [nc sort $node] }
	}
	delayed [nc sort root]
}

proc SortAlias {name val} {
	set ns [string totitle [string range $name 0 end-1]]
	set nc [namespace current]
	interp alias {} ${nc}::sort$ns {} ${nc}::sort::${name}::$val
}

proc DoSort {uid} {
	variable hidden
	variable mapping
	if {[info exists hidden($uid)] || ![info exists mapping($uid)]} return
	foreach x $mapping($uid) { delayed [nc sort [NodeParent $x]] }
}

set all_node_options [list]
foreach x [info commands node-*] { 
	lappend all_node_options [string range $x 4 end]
}

#Handle tree nodes mapping changes
proc Mapping {uid args} {
	variable mapping
	variable hidden
	set new [mapper $uid]
	if {![info exists hidden($uid)]} {
		if {[info exists mapping($uid)]} {
			if {$mapping($uid)==$new} return
			foreach {_ show hide}\
				[lcompare $mapping($uid) $new] break
			foreach x $hide { do_hide $x }
		} else { set show $new }
		foreach x $show { do_show $x $uid }
	}
	set mapping($uid) $new
}

proc MapChildren {uid args} {
	variable mapping
	Mapping $uid
	set g [lindex [split $uid :] end]
	foreach x [select Contact "\[lsearch \$Groups $g\]!=-1"] {
		if {[info exists mapping($x)]} { Mapping $x }
	}
	foreach x [select Group:common "\[lsearch \$Groups $g\]!=-1"] {
		if {[info exists mapping($x)]} { MapChildren $x }
	}
}

proc mapper {uid} {
	variable mapping
	set lst [list]
	foreach parent [grouper $uid] {
		if {[info exists mapping($parent)]} {
			foreach y $mapping($parent) { lappend lst "$y\0$uid" }
		} else { set root 1 }
	}
	if {[info exists root] && ![llength $lst]} { set lst "root\0$uid" }
	set lst
}

proc trigger_offline {args} {
	variable old_status
	foreach x [select Contact] {
		if {![info exists old_status($x)]} { hide $x $hide::offline }
	}
}

proc trigger {name} { set $name [expr [string is true [set $name]]^1] }

proc UpdateStatus {args} {
	upvar #0 [ref Me](Status) status
	.status configure -image img:$status -text [mc $status]
	ui::tooltip::action .status "[mc {Change status}]\n[mc {Now status is}] [mc $status]"
	WindowTitle
}
proc WindowTitle {args} {
	wm title . "Alicq: [get Me Alias] [get Me Status]"
}

# Menu handling 
proc MakeMenu {menu {object ""} {context ""}} {
	if {[winfo exists $menu]} { destroy $menu }
	menu $menu -type normal -tearoff no
	if {$object!=""} {
		set r [list [ref $object]]
		while {[lindex $r end]!="::objects"} {
			lappend r [namespace qualifier [lindex $r end]]
		}
	} else { set r [namespace children ::modules] }
	NewMenu $menu $r $object
}

proc getW {n} {
	set n [lindex [namespace tail [lindex $n 1]] 0]
	if {[string is double $n] && $n >=0 && $n <= 1} { return $n }
	return 0.5
}
proc cmp2 {n1 n2} {
	if {[set w1 [getW $n1]]==[set w2 [getW $n2]]} {
		string compare [namespace tail $n1] [namespace tail $n2]
	} else { expr {($w1>$w2)?1:(($w1<$w2)?-1:0)} }
}

proc menu-commands {menu cmd obj} { 
	set cmd [expr { ($obj!="")?[list $cmd $obj]:[list $cmd] }]
	$menu add command -command $cmd 
	return 1
}

proc menu-vars {menu var obj} {
	if {$obj==""} {
		set var [namespace parent [namespace parent\
			[namespace qualifier $var]]]::[set $var]
		upvar #0 [var2meta $var] meta
	} else {
		upvar #0 [namespace parent\
			[namespace qualifier $var]]::[set $var] meta
		set var [ref $obj]([set $var])
		if {![info exists $var] && [info exists meta(default)]} {
			set $var $meta(default)
		}
	}
	if {![info exists meta]} { return 0 }
	if {[info exists meta(type)] && [info commands menu-$meta(type)]!=""} {
		menu-$meta(type) $menu $var $obj
	} else { return 0 }
}

proc menu-variant {menu var obj} {
	upvar 1 meta meta
	foreach x {valuescript values} { 
		if {[info exists meta($x)]} { set opt $x; break }
	}
	if {![info exists opt]} { return 0 }
	set n $menu.mn[clock seconds]-[clock clicks]
	$menu add cascade -menu\
		[ui::variant $n -variable $var -$opt $meta($opt)]
	return 1
}

proc menu-boolean {menu var object} {
	$menu add checkbutton -variable $var
	return 1
}

proc NewMenu {menu nslist {obj ""}} {
	set res [list]
	foreach ns $nslist {
		# Add menu commands
		foreach what {commands vars} {
			foreach x [info $what ${ns}::meta::menu::*] {
				lappend res [list $what $x]
			}
		}
	}
	foreach item [lsort -command cmp2 $res] {
		foreach {cmd data} $item break
		set name [namespace tail $data]
		if {[string is double [lindex $name 0]]} {
			set name [lrange $name 1 end]
		}
		set m [ui::subcascade $menu [lrange $name 0 end-1]]	
		if {[menu-$cmd $m $data $obj]} {
			$m entryconfigure last -label [mc [lindex $name end]]
		}
	}
}

init
